//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Views/ImportDataWidgets/RealDataSelectorWidget.cpp
//! @brief     Implements class RealDataSelectorWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Views/ImportDataWidgets/RealDataSelectorWidget.h"
#include "Device/Data/DataUtils.h"
#include "GUI/Application/Application.h"
#include "GUI/Models/DataLoaders1D.h"
#include "GUI/Models/Error.h"
#include "GUI/Models/RealDataItem.h"
#include "GUI/Views/CommonWidgets/GUIHelpers.h"
#include "GUI/Views/CommonWidgets/StyledToolBar.h"
#include "GUI/Views/ImportDataWidgets/ImportDataUtils.h"
#include "GUI/Views/ImportDataWidgets/RealDataPropertiesWidget.h"
#include "GUI/Views/ImportDataWidgets/RealDataTreeModel.h"
#include "GUI/mainwindow/ProjectManagerStore.h"
#include "GUI/mainwindow/mainwindow.h"
#include "GUI/mainwindow/projectmanager.h"
#include <QApplication>
#include <QFileDialog>
#include <QItemSelectionModel>
#include <QLineEdit>
#include <QListView>
#include <QMenu>
#include <QSplitter>
#include <QTreeView>
#include <QVBoxLayout>

RealDataSelectorWidget::RealDataSelectorWidget(QWidget* parent)
    : QWidget(parent)
    , m_itemTree(new QTreeView)
    , m_itemTreeModel(new RealDataTreeModel)
    , m_propertiesWidget(new RealDataPropertiesWidget)
    , m_import2dDataAction(new QAction(this))
    , m_import1dDataAction(new QAction(this))
    , m_renameDataAction(new QAction(this))
    , m_removeDataAction(new QAction(this))
    , m_rotateDataAction(new QAction(this))
{
    setMinimumSize(128, 600);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setWindowTitle("RealDataSelectorWidget");

    // #baimport ++ necessary to tree somehow?
    /*
    QSize RealDataItemSelectorWidget::sizeHint() const
    {
        return QSize(GUI::Constants::ITEM_SELECTOR_WIDGET_WIDTH,
    GUI::Constants::ITEM_SELECTOR_WIDGET_HEIGHT);
    }

    QSize RealDataItemSelectorWidget::minimumSizeHint() const { return QSize(25, 25); }
    */

    m_import2dDataAction->setText("Import 2D data");
    m_import2dDataAction->setIcon(QIcon(":/images/import.svg"));
    m_import2dDataAction->setIconText("2D");
    m_import2dDataAction->setToolTip("Import 2D data");
    connect(m_import2dDataAction, &QAction::triggered, [this]() { importData2D(); });

    m_import1dDataAction->setText("Import 1D data");
    m_import1dDataAction->setIcon(QIcon(":/images/import.svg"));
    m_import1dDataAction->setIconText("1D");
    m_import1dDataAction->setToolTip("Import 1D data");
    connect(m_import1dDataAction, &QAction::triggered, [this]() { importData1D(); });

    m_renameDataAction->setText("Rename");
    m_renameDataAction->setIcon(QIcon()); // #baTODO: Icon needed?
    m_renameDataAction->setIconText("Rename");
    m_renameDataAction->setToolTip("Rename data");
    connect(m_renameDataAction, &QAction::triggered, this,
            &RealDataSelectorWidget::renameCurrentItem);

    m_removeDataAction->setText("Remove");
    m_removeDataAction->setIcon(QIcon(":/images/delete.svg"));
    m_removeDataAction->setIconText("Remove");
    m_removeDataAction->setToolTip("Remove selected data");
    connect(m_removeDataAction, &QAction::triggered, this,
            &RealDataSelectorWidget::removeCurrentItem);

    m_rotateDataAction->setText("Rotate");
    m_rotateDataAction->setIcon(QIcon(":/images/rotate-left.svg"));
    m_rotateDataAction->setIconText("Rotate");
    m_rotateDataAction->setToolTip("Rotate intensity data by 90 deg counterclockwise");
    connect(m_rotateDataAction, &QAction::triggered, this,
            &RealDataSelectorWidget::rotateCurrentItem);

    QToolBar* toolBar = new StyledToolBar(this);
    toolBar->setMinimumSize(toolBar->minimumHeight(), toolBar->minimumHeight());
    toolBar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    toolBar->addAction(m_import2dDataAction);
    toolBar->addAction(m_import1dDataAction);
    toolBar->addAction(m_removeDataAction);
    toolBar->addAction(m_rotateDataAction);

    // #baimport ++ following line necessary for tree as well? relevant for RealItems at all?
    // m_itemTree->setAttribute(Qt::WA_MacShowFocusRect, false);
    m_itemTree->setItemsExpandable(false);
    m_itemTree->setRootIsDecorated(false);
    m_itemTree->setHeaderHidden(true);
    m_itemTree->setContextMenuPolicy(Qt::CustomContextMenu);
    m_itemTree->setModel(m_itemTreeModel);

    auto splitter = new QSplitter;
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_itemTree);
    splitter->addWidget(m_propertiesWidget);
    splitter->setChildrenCollapsible(true);

    auto mainLayout = new QVBoxLayout;
    mainLayout->setMargin(0);
    mainLayout->setSpacing(0);
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->addWidget(toolBar);
    mainLayout->addWidget(splitter);
    setLayout(mainLayout);

    connect(m_itemTree, &QTreeView::customContextMenuRequested, this,
            &RealDataSelectorWidget::onContextMenuRequest);

    updateActionEnabling();
}

QSize RealDataSelectorWidget::sizeHint() const
{
    return QSize(200, 400);
}

QSize RealDataSelectorWidget::minimumSizeHint() const
{
    return QSize(128, 200);
}

void RealDataSelectorWidget::setModel(RealDataModel* realDataModel)
{
    m_itemTreeModel->setRealDataModel(realDataModel);

    connect(m_itemTreeModel, &QAbstractItemModel::modelReset,
            [this]() { m_itemTree->expandAll(); });

    connect(m_itemTreeModel, &QAbstractItemModel::rowsInserted,
            [this]() { m_itemTree->expandAll(); });

    connect(m_itemTree->selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &RealDataSelectorWidget::onSelectionChanged, Qt::UniqueConnection);
}

RealDataItem* RealDataSelectorWidget::currentItem()
{
    return m_itemTreeModel->itemForIndex(currentIndex());
}

void RealDataSelectorWidget::setCurrentItem(RealDataItem* item)
{
    m_itemTree->selectionModel()->clearSelection();
    QModelIndex index = m_itemTreeModel->indexForItem(item);
    if (index.isValid())
        m_itemTree->selectionModel()->setCurrentIndex(index, QItemSelectionModel::SelectCurrent);
}

QModelIndex RealDataSelectorWidget::currentIndex()
{
    return m_itemTree->selectionModel()->currentIndex();
}

void RealDataSelectorWidget::onSelectionChanged()
{
    updateActionEnabling();
    m_propertiesWidget->setItem(currentItem());
    emit selectionChanged(currentItem());
}

void RealDataSelectorWidget::onContextMenuRequest(const QPoint& point)
{
    auto realDataItemAtPoint = m_itemTreeModel->itemForIndex(m_itemTree->indexAt(point));
    updateActionEnabling(realDataItemAtPoint);

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (realDataItemAtPoint != nullptr) {
        menu.addAction(m_renameDataAction);
        menu.addAction(m_removeDataAction);
        if (realDataItemAtPoint->isIntensityData())
            menu.addAction(m_rotateDataAction);
        menu.addSeparator();
    }

    menu.addAction(m_import2dDataAction);
    menu.addAction(m_import1dDataAction);
    menu.exec(m_itemTree->mapToGlobal(point));
}

void RealDataSelectorWidget::importData1D()
{
    QMap<QString, AbstractDataLoader*> loaderOfFilter;
    QString filters;
    for (auto loader : DataLoaders1D::instance().loaders()) {
        const QString filter =
            loader->name() + " (*.txt *.csv *.dat)"; // #baimport - take file filters from loader
        loaderOfFilter[filter] = loader;

        if (!filters.isEmpty())
            filters += ";;";
        filters += filter;
    }
    filters += ";;Other (*.*)";

    QString selectedFilter = GUI::Project::Manager::projectManager()->recentlyUsedImportFilter1D();
    const QString dirname = GUI::Project::Manager::projectManager()->userImportDir();

    const QStringList fileNames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        baApp->settings().useNativeFileDialog() ? QFileDialog::Options()
                                                : QFileDialog::DontUseNativeDialog);

    if (fileNames.isEmpty())
        return;

    GUI::Project::Manager::projectManager()->setImportDirFromFilePath(fileNames[0]);
    GUI::Project::Manager::projectManager()->setRecentlyUsedImportFilter1D(selectedFilter);

    const AbstractDataLoader* selectedLoader = loaderOfFilter.value(selectedFilter, nullptr);
    for (const auto& fileName : fileNames) {
        RealDataItem* realDataItem = m_itemTreeModel->insertSpecularDataItem();
        realDataItem->setName(QFileInfo(fileName).baseName());
        realDataItem->setNativeFileName(fileName);

        const QString errorText =
            GUI::View::ImportDataUtils::Import1dData(realDataItem, selectedLoader);
        if (errorText.isEmpty())
            setCurrentItem(realDataItem);
        else {
            m_itemTreeModel->removeItem(realDataItem);
            GUI::View::Helpers::warning(this, "File import",
                                        QString("The file '%1' could not be imported.")
                                            .arg(QDir::toNativeSeparators(fileName)),
                                        errorText);
        }
    }
}

void RealDataSelectorWidget::importData2D()
{
    QMap<QString, IntensityDataIOFactory::LoaderSelector> loaderOfFilter;
    QString filters;

    const auto addFilter = [&](IntensityDataIOFactory::LoaderSelector loaderSelector,
                               const QString& filter) {
        if (!filters.isEmpty())
            filters += ";;";
        filters += filter;
        loaderOfFilter[filter] = loaderSelector;
    };
    addFilter(IntensityDataIOFactory::tiff, "TIFF (*.tif *.tiff *.tif.gz)");
    addFilter(IntensityDataIOFactory::nicos, "Nicos (*.001)");
    addFilter(IntensityDataIOFactory::bornagain, "BornAgain (*.int.gz)");
    addFilter(IntensityDataIOFactory::automatic, "CSV (*.txt *.csv *.dat)");
    addFilter(IntensityDataIOFactory::automatic, "All (*.*)");

    QString selectedFilter = GUI::Project::Manager::projectManager()->recentlyUsedImportFilter2D();

    const QString dirname = GUI::Project::Manager::projectManager()->userImportDir();
    const QStringList fileNames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        baApp->settings().useNativeFileDialog() ? QFileDialog::Options()
                                                : QFileDialog::DontUseNativeDialog);
    if (fileNames.isEmpty())
        return;

    GUI::Project::Manager::projectManager()->setImportDirFromFilePath(fileNames[0]);
    GUI::Project::Manager::projectManager()->setRecentlyUsedImportFilter2D(selectedFilter);

    const auto selectedLoader =
        loaderOfFilter.value(selectedFilter, IntensityDataIOFactory::automatic);

    for (const auto& fileName : fileNames) {
        std::unique_ptr<OutputData<double>> data =
            GUI::View::ImportDataUtils::Import2dData(fileName, selectedLoader);
        if (!data)
            continue;
        RealDataItem* realDataItem = m_itemTreeModel->insertIntensityDataItem();
        realDataItem->setName(QFileInfo(fileName).baseName());

        try {
            if (data->rank() != 2)
                throw Error("The content could not be interpreted correctly as 2-dimensional.");

            realDataItem->setOutputData(data.release());
            setCurrentItem(realDataItem);
        } catch (std::exception& ex) {
            m_itemTreeModel->removeItem(realDataItem);

            const QString message = QString("Error while trying to read file\n\n'%1'\n\n%2")
                                        .arg(fileName)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::View::Helpers::warning(this, "File import", message);
        }
    }
}

void RealDataSelectorWidget::renameCurrentItem()
{
    if (currentIndex().isValid())
        m_itemTree->edit(currentIndex());
}

void RealDataSelectorWidget::removeCurrentItem()
{
    m_itemTreeModel->removeItem(currentItem());
}

void RealDataSelectorWidget::rotateCurrentItem()
{
    if (!currentItem() || !currentItem()->isIntensityData())
        // should never happen because of action disabling => no
        // dialog necessary
        return;

    if (currentItem()->rotationAffectsSetup()) {
        const QString title("Rotate data");
        const QString message("Rotation will break the link between the data and the instrument. "
                              "Detector masks or projections, if they exist, will be removed.");
        if (!GUI::View::Helpers::question(MainWindow::instance(), title, message,
                                          "Do you wish to rotate the data?", "Yes, please rotate",
                                          "No, cancel data rotation"))
            return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    currentItem()->rotateData();
    QApplication::restoreOverrideCursor();
}

void RealDataSelectorWidget::updateActionEnabling()
{
    updateActionEnabling(currentItem());
}

void RealDataSelectorWidget::updateActionEnabling(const RealDataItem* item)
{
    m_import2dDataAction->setEnabled(true);
    m_import1dDataAction->setEnabled(true);

    m_rotateDataAction->setEnabled(item != nullptr ? item->isIntensityData() : false);
    m_removeDataAction->setEnabled(item != nullptr);
    m_renameDataAction->setEnabled(item != nullptr);
}

void RealDataSelectorWidget::showEvent(QShowEvent*)
{
    // ensure a current item when widget is shown
    if (!currentItem())
        setCurrentItem(m_itemTreeModel->topMostItem());
}
