//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/DWBA/DWBA_Scalar.cpp
//! @brief     Implements class DWBA_Scalar.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/DWBA/DWBA_Scalar.h"
#include "Base/Vector/WavevectorInfo.h"
#include "Resample/Element/DiffuseElement.h"
#include "Resample/Flux/ScalarFlux.h"
#include "Sample/Scattering/IFormFactor.h"

DWBA_Scalar::DWBA_Scalar(const IFormFactor& ff, size_t i_layer) : IDWBA(ff, i_layer) {}
DWBA_Scalar::DWBA_Scalar(const IFormFactor& ff) : IDWBA(ff) {}

DWBA_Scalar::~DWBA_Scalar() = default;

complex_t DWBA_Scalar::coherentFF(const DiffuseElement& ele) const
{
    const WavevectorInfo& wavevectors = ele.wavevectorInfo();

    if (!m_i_layer.has_value())
        // no slicing, pure Born approximation
        return m_ff->theFF(wavevectors);

    const auto* inFlux = dynamic_cast<const ScalarFlux*>(ele.fluxIn(iLayer()));
    const auto* outFlux = dynamic_cast<const ScalarFlux*>(ele.fluxOut(iLayer()));

    // Retrieve the two different incoming wavevectors in the layer
    const cvector_t& ki = wavevectors.getKi();
    const complex_t kiz = inFlux->getScalarKz();
    const cvector_t k_i_T{ki.x(), ki.y(), -kiz};
    const cvector_t k_i_R{ki.x(), ki.y(), +kiz};

    // Retrieve the two different outgoing wavevector bins in the layer
    const cvector_t& kf = wavevectors.getKf();
    const complex_t kfz = outFlux->getScalarKz();
    const cvector_t k_f_T{kf.x(), kf.y(), +kfz};
    const cvector_t k_f_R{kf.x(), kf.y(), -kfz};

    // Construct the four different scattering contributions wavevector infos
    const double wavelength = wavevectors.vacuumLambda();
    const WavevectorInfo k_TT(k_i_T, k_f_T, wavelength);
    const WavevectorInfo k_RT(k_i_R, k_f_T, wavelength);
    const WavevectorInfo k_TR(k_i_T, k_f_R, wavelength);
    const WavevectorInfo k_RR(k_i_R, k_f_R, wavelength);

    // Get the four R,T coefficients
    const complex_t T_in = inFlux->getScalarT();
    const complex_t R_in = inFlux->getScalarR();
    const complex_t T_out = outFlux->getScalarT();
    const complex_t R_out = outFlux->getScalarR();

    // The four different scattering contributions;
    // S stands for scattering off the particle,
    // R for reflection off the layer interface.
    const complex_t term_S = T_in * m_ff->theFF(k_TT) * T_out;
    const complex_t term_RS = R_in * m_ff->theFF(k_RT) * T_out;
    const complex_t term_SR = T_in * m_ff->theFF(k_TR) * R_out;
    const complex_t term_RSR = R_in * m_ff->theFF(k_RR) * R_out;

    return term_S + term_RS + term_SR + term_RSR;
}
