//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Specular/SpecularMagneticStrategy.h
//! @brief     Defines class SpecularMagneticStrategy.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2020
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_RESAMPLE_SPECULAR_SPECULARMAGNETICSTRATEGY_H
#define BORNAGAIN_RESAMPLE_SPECULAR_SPECULARMAGNETICSTRATEGY_H

#include "Resample/Flux/MatrixFlux.h"
#include "Resample/Specular/ISpecularStrategy.h"
#include <memory>
#include <vector>

class SliceStack;

//! Implements the magnetic Fresnel computation with Nevot-Croce roughness
//!
//! Implements the transfer matrix formalism for the calculation of wave
//! amplitudes of the coherent wave solution in a multilayer with magnetization.
//! For a description, see internal
//! document "Polarized Implementation of the Transfer Matrix Method"
class SpecularMagneticStrategy : public ISpecularStrategy {
public:
    //! Computes refraction angle reflection/transmission coefficients
    //! for given sliced multilayer and wavevector k
    Fluxes Execute(const SliceStack& slices, const kvector_t& k) const override;

    //! Computes refraction angle reflection/transmission coefficients
    //! for given sliced multilayer and a set of kz projections corresponding to each slice
    Fluxes Execute(const SliceStack& slices, const std::vector<complex_t>& kz) const override;

    //! Computes the Fresnel R coefficient for the top layer only
    //! Introduced in order to speed up pure reflectivity computations
    virtual std::variant<complex_t, Eigen::Matrix2cd>
    computeTopLayerR(const SliceStack& slices, const std::vector<complex_t>& kzs) const override;

private:
    std::vector<MatrixFlux> computeFlux(const SliceStack& slices,
                                        const std::vector<complex_t>& kzs) const;

    virtual std::pair<Eigen::Matrix2cd, Eigen::Matrix2cd>
    computeBackwardsSubmatrices(const MatrixFlux& coeff_i, const MatrixFlux& coeff_i1,
                                double sigma) const = 0;

    void calculateUpwards(std::vector<MatrixFlux>& coeff, const SliceStack& slices) const;
};

#endif // BORNAGAIN_RESAMPLE_SPECULAR_SPECULARMAGNETICSTRATEGY_H
#endif // USER_API
